#ifndef ICINGADB_TI
#define ICINGADB_TI

#include "base/object.hpp"
#include "base/type.hpp"
#include "base/value.hpp"
#include "base/array.hpp"
#include "base/atomic.hpp"
#include "base/dictionary.hpp"
#include <boost/signals2.hpp>

#include "base/configobject.hpp"

namespace icinga
{

class IcingaDB;

template<>
class TypeImpl<IcingaDB> : public TypeImpl<ConfigObject>
{
public:
	DECLARE_PTR_TYPEDEFS(TypeImpl<IcingaDB>);

	TypeImpl();
	~TypeImpl() override;

	String GetName() const override;
	int GetAttributes() const override;
	Type::Ptr GetBaseType() const override;
	int GetFieldId(const String& name) const override;
	Field GetFieldInfo(int id) const override;
	int GetFieldCount() const override;
	ObjectFactory GetFactory() const override;
	std::vector<String> GetLoadDependencies() const override;
	int GetActivationPriority() const override;
public:
	void RegisterAttributeHandler(int fieldId, const Type::AttributeHandler& callback) override;
};


template<>
class ObjectImpl<IcingaDB> : public ConfigObject
{
public:
	DECLARE_PTR_TYPEDEFS(ObjectImpl<IcingaDB>);

	void Validate(int types, const ValidationUtils& utils) override;
	void SimpleValidateHost(const Lazy<String>& value, const ValidationUtils& utils);
	void SimpleValidatePath(const Lazy<String>& value, const ValidationUtils& utils);
	void SimpleValidatePassword(const Lazy<String>& value, const ValidationUtils& utils);
	void SimpleValidatePort(const Lazy<int>& value, const ValidationUtils& utils);
	void SimpleValidateDbIndex(const Lazy<int>& value, const ValidationUtils& utils);
public:
	ObjectImpl<IcingaDB>();
public:
	~ObjectImpl<IcingaDB>() override;
public:
	void SetField(int id, const Value& value, bool suppress_events = false, const Value& cookie = Empty) override;
public:
	Value GetField(int id) const override;
public:
	void ValidateField(int id, const Lazy<Value>& lvalue, const ValidationUtils& utils) override;
public:
	void NotifyField(int id, const Value& cookie = Empty) override;
public:
	Object::Ptr NavigateField(int id) const override;
public:
	String GetHost() const;
public:
	String GetPath() const;
public:
	String GetPassword() const;
public:
	int GetPort() const;
public:
	int GetDbIndex() const;
public:
	void SetHost(const String& value, bool suppress_events = false, const Value& cookie = Empty);
public:
	void SetPath(const String& value, bool suppress_events = false, const Value& cookie = Empty);
public:
	void SetPassword(const String& value, bool suppress_events = false, const Value& cookie = Empty);
public:
	void SetPort(int value, bool suppress_events = false, const Value& cookie = Empty);
public:
	void SetDbIndex(int value, bool suppress_events = false, const Value& cookie = Empty);
protected:
public:
	virtual void NotifyHost(const Value& cookie = Empty);
public:
	virtual void NotifyPath(const Value& cookie = Empty);
public:
	virtual void NotifyPassword(const Value& cookie = Empty);
public:
	virtual void NotifyPort(const Value& cookie = Empty);
public:
	virtual void NotifyDbIndex(const Value& cookie = Empty);
private:
	inline String GetDefaultHost() const;
private:
	inline String GetDefaultPath() const;
private:
	inline String GetDefaultPassword() const;
private:
	inline int GetDefaultPort() const;
private:
	inline int GetDefaultDbIndex() const;
protected:
	virtual void ValidateHost(const Lazy<String>& lvalue, const ValidationUtils& utils);
protected:
	virtual void ValidatePath(const Lazy<String>& lvalue, const ValidationUtils& utils);
protected:
	virtual void ValidatePassword(const Lazy<String>& lvalue, const ValidationUtils& utils);
protected:
	virtual void ValidatePort(const Lazy<int>& lvalue, const ValidationUtils& utils);
protected:
	virtual void ValidateDbIndex(const Lazy<int>& lvalue, const ValidationUtils& utils);
private:
	EventuallyAtomic<String>::type m_Host;
	EventuallyAtomic<String>::type m_Path;
	EventuallyAtomic<String>::type m_Password;
	EventuallyAtomic<int>::type m_Port;
	EventuallyAtomic<int>::type m_DbIndex;
public:
	static boost::signals2::signal<void (const intrusive_ptr<IcingaDB>&, const Value&)> OnHostChanged;
	static boost::signals2::signal<void (const intrusive_ptr<IcingaDB>&, const Value&)> OnPathChanged;
	static boost::signals2::signal<void (const intrusive_ptr<IcingaDB>&, const Value&)> OnPasswordChanged;
	static boost::signals2::signal<void (const intrusive_ptr<IcingaDB>&, const Value&)> OnPortChanged;
	static boost::signals2::signal<void (const intrusive_ptr<IcingaDB>&, const Value&)> OnDbIndexChanged;
};

}
#endif /* ICINGADB_TI */
